/**
* p2pwnc-tc.c - P2PWNC Traffic Controller module
*
* Copyright (C) 2005-2006 Vasileios P. Kemerlis
*	
* Author:  Vasileios P. Kemerlis
* Contact: vpk@cs.aueb.gr
*          hdkiller@venus.cslab.aueb.gr
*          http://www.mm.aueb.gr/~vpk/
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*
*/

#include "p2pwnc-tc.h"

//################################################################
/* 
 * Memory deallocator 
 */
void dealloc(void) {
	//use global variables 
	extern config_file cf;
	extern FILE* pidfile;
	
	//clear configuration file allocations first
	if(cf.path != NULL) {
		free(cf.path);
		cf.path = NULL;
	}
	
	if(cf.inet_dev != NULL) {
		free(cf.inet_dev);
		cf.inet_dev = NULL;
	}
	
	if(cf.p2pwnc_dev != NULL) {
		free(cf.p2pwnc_dev);
		cf.p2pwnc_dev = NULL;
	}
	
	if(cf.tc != NULL) {
		free(cf.tc);
		cf.tc = NULL;
	}
	
	if(cf.iptables != NULL) {
		free(cf.iptables);
		cf.tc = NULL;
	}
	
	if(cf.cless_qdisc_opt != NULL) {
		free(cf.cless_qdisc_opt);
		cf.cless_qdisc_opt = NULL;
	}

	//clean resource pool 
	p2pwnc_client *clp, *deallocp; //pointers to traverse the list
	clp = rpool.cls;
	deallocp = NULL;
	while(clp != NULL){ //traverse all records and deallocate 
		deallocp = clp;
		clp = (p2pwnc_client*)deallocp->next;
		free(deallocp);
		deallocp = NULL;
	}
	
	if(pidfilename != NULL) {
		free(pidfilename);
		pidfilename = NULL;
	}
	
}
//################################################################

//################################################################
/*
 * timer startup/shutdown
 * set the appropriate signal handlers
 * start - stop the timer
 * action = 0 activate the timer
 * action = 1 reset the timer
 */
void set_timer(struct sigaction* alrm, uint8_t action) {
	switch(action) {
		case ACTIVATE_TIMER: //init timer - set the alarm handler etc...
			(void)sigfillset(&(*alrm).sa_mask); //block all signals when alarm signal handler run 
			alrm->sa_flags = 0;
			alrm->sa_handler = close_connection;
			alarm_sig = 0;
			(void)sigaction(SIGALRM, alrm, NULL);
			(void)alarm(cf.timeout);
			break;
		case DEACTIVATE_TIMER: //deactivate timer
			(void)sigemptyset(&(*alrm).sa_mask); 
			alrm->sa_flags = 0;
			alrm->sa_handler = SIG_IGN; // ignore the signal
			alarm_sig = 0;
			(void)sigaction(SIGALRM, alrm, NULL);
			break;
	}

}

//################################################################

//################################################################
/*
 * Print logging info to syslog for the active reservations
 */
void log_updated_tc(void) {
	
	p2pwnc_client *clp; //pointer to traverse the list
	clp = rpool.cls;
	
	syslog(LOG_FACILITY | LOG_INFO, "Start of resource-reservations list");
	syslog(LOG_FACILITY | LOG_INFO, "Reservations: %d, QoS-Level sum: %d", rpool.num, rpool.qos_level_sum);
	while(clp != NULL){ //traverse all records 
		syslog(LOG_FACILITY | LOG_INFO, "Client: %s, QoS-Level: %d, Uplink: %d, Downlink: %d", inet_ntoa(clp->ip_address), clp->qos_level, (clp->qos_level * cf.puplink) / rpool.qos_level_sum, (clp->qos_level * cf.pdownlink) / rpool.qos_level_sum);
		clp = (p2pwnc_client*)clp->next;
	}
	syslog(LOG_FACILITY | LOG_INFO, "End of resource-reservations list");
	
}
//################################################################

//################################################################
/*
 * update resource pool
 * if qlevel = 0 then remove address from reservation pool and 
 * 	update the appropriate structures
 * 	
 * if qlevel != 0 then update (or add a new entry) the reservation pool
 * 	and update the appropriate structures
 * 	
 * if client is a team member team mode is enabled then update is ignored
 */
uint8_t update_resource_pool(struct in_addr address, uint8_t team, uint8_t qlevel) {
	
	p2pwnc_client *clp, *clpp;
	clp = rpool.cls;
	clpp = NULL;
		
	//if p2pwnc client and home user belong to the same team
	//and team mode is enabled we don't allocate anything for him
	if(team && cf.team)
		return 1;
	
	//p2pwnc client is not member of home user's team or 
	//team mode is dissabled
	if(!qlevel) { //remove - if qos_level = 0 then remove the p2pwnc client from the resource pool
		
		while(clp != NULL && clp->ip_address.s_addr != address.s_addr) {//traverse to find which client is to remove
			clpp = clp;//pointer to the previous client
			clp = (p2pwnc_client*)clp->next;
		}
		
		if((clp != NULL) && (clp->ip_address.s_addr == address.s_addr)) {//address found - remove
			//rearrange list
			if(clpp == NULL) //previous is null = remove the list head
				rpool.cls = clp->next;
			else
				clpp->next = clp->next;
			//update qos sum and client counter
			rpool.qos_level_sum -= clp->qos_level;
			rpool.num--;
			//deallocate
			free(clp);
			
		}
		else
			return 0; //unable to remove - invalid address
	}
	else {//add the new client to the list
		
		//the list has some elements
		if(clp != NULL) { 
			while(clp->next != NULL && clp->ip_address.s_addr != address.s_addr) //traverse to find where add
				clp = (p2pwnc_client*)clp->next;
			if(clp->ip_address.s_addr != address.s_addr ) {//add a new client to the end
				//memory allocation
				clpp = (p2pwnc_client*)malloc(sizeof(p2pwnc_client));
				if(clpp == NULL) //fail
					fail(strerror(errno),1);
				(void)memset(clpp, 0, sizeof(p2pwnc_client));
				clpp->next = NULL;
				//connect to the end
				clp->next = clpp;
				clp = clpp;
				//update client counter
				rpool.num++;
			}
			//update an existing or the last entry
			clp->ip_address = address;
			clp->team_member= team;
			//remove the old qos_level from the sum
			rpool.qos_level_sum -= clp->qos_level;	
			//udate qos_level and add to the sum
			clp->qos_level = qlevel;
			rpool.qos_level_sum += clp->qos_level;	
		}
		else { //the list is empty
			//memory allocation
			clp = (p2pwnc_client*)malloc(sizeof(p2pwnc_client));
			if(clp == NULL) //fail
				fail(strerror(errno),1);
			(void)memset(clp, 0, sizeof(p2pwnc_client));
			//update new entry
			clp->ip_address = address;
			clp->team_member = team;
			clp->qos_level = qlevel;
			clp->next = NULL;
			//connect to the beginning
			rpool.cls = clp;
			//update clients number
			rpool.num++;
			rpool.qos_level_sum += clp->qos_level;
		}
	}
	return 1;
}
//################################################################

//################################################################
/*
 * initialize tc 
 * mode = 0 -> clear tc qdiscs - iptables rules
 * mode = 1 -> clear tc qdiscs - iptables rules 
 * 	and initialize qdiscs for all reservation queries
 */

uint8_t init_tc(uint8_t mode, struct in_addr* address) {
	char cmdbuf[2*MSGBUF_MAX]; //command exec buffer
	p2pwnc_client *clp; //traverse allocation pool
	int i; //general purpose counter
	size_t max, min, limit, burst; //RED qdisc parameters
	FILE* fp = fopen(TMPFILE,"w"); //the output tc script
	
	//clean tc
	switch(cf.cfull_qdisc) { 
		case HTB: //htb handle	
			//inet_dev
			(void)fprintf(fp, "%s qdisc del dev %s root > /dev/null 2>&1\n", cf.tc, cf.inet_dev);
			//p2pwnc_dev
			(void)fprintf(fp, "%s qdisc del dev %s root > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev);
			//clean iptables
			(void)fprintf(fp, "%s -t mangle -F PREROUTING > /dev/null 2>&1\n", cf.iptables);
			(void)fprintf(fp, "%s -t mangle -F POSTROUTING > /dev/null 2>&1\n", cf.iptables);
		break;
	}	
	if(mode == CLEAR){ //clear only no init
		(void)fclose(fp);//close the script file
		(void)memset(cmdbuf, 0, sizeof(cmdbuf));//generate the exec command
		(void)sprintf(cmdbuf, "/bin/sh %s",TMPFILE);
		if(system(cmdbuf) == -1) {//exec
			(void)unlink(TMPFILE);//remove tc script
			return 0;
		}
		else {
			(void)unlink(TMPFILE);//remove tc script
			return 1;
		}
	}
	else { //clear and init - clean was done before so we only have to init tc and iptables
		
		switch(cf.cfull_qdisc) { 
			case HTB: //htb qdisc
			
			if(cf.cfull_qdisc_opt != NULL && cf.cfull_qdisc_opt->htb_r2q != 0) {
				
				//root handle for inet_dev
				(void)fprintf(fp, "%s qdisc add dev %s root handle 1: htb default 2 r2q %d > /dev/null 2>&1\n", cf.tc, cf.inet_dev, cf.cfull_qdisc_opt->htb_r2q);
				//root handle for p2pwnc_dev
				(void)fprintf(fp, "%s qdisc add dev %s root handle 1: htb default 2 r2q %d > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, cf.cfull_qdisc_opt->htb_r2q);
			}
			else {
				//root handle for inet_dev
				(void)fprintf(fp, "%s qdisc add dev %s root handle 1: htb default 2 > /dev/null 2>&1\n", cf.tc, cf.inet_dev);
				//root handle for p2pwnc_dev
				(void)fprintf(fp, "%s qdisc add dev %s root handle 1: htb default 2 > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev);
			}
			
			if(cf.cfull_qdisc_opt != NULL && cf.cfull_qdisc_opt->htb_quantum != 0) {
				//full bw class for inet_dev - UPLINK
				(void)fprintf(fp, "%s class add dev %s parent 1:0 classid 1:1 htb rate %dkbit ceil %dkbit quantum %d > /dev/null 2>&1\n", cf.tc, cf.inet_dev, cf.ruplink, cf.ruplink, cf.cfull_qdisc_opt->htb_quantum);
				//full bw class for p2pwnc_dev - DOWNLINK
				(void)fprintf(fp, "%s class add dev %s parent 1:0 classid 1:1 htb rate %dkbit ceil %dkbit quantum %d > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, cf.rdownlink, cf.rdownlink, cf.cfull_qdisc_opt->htb_quantum);

				//home bw class for inet_dev - UPLINK
				(void)fprintf(fp, "%s class add dev %s parent 1:1 classid 1:2 htb rate %dkbit ceil %dkbit quantum %d > /dev/null 2>&1\n", cf.tc, cf.inet_dev, cf.ruplink - cf.puplink, cf.ruplink, cf.cfull_qdisc_opt->htb_quantum);
				//home bw class for p2pwnc_dev - DOWNLINK
				(void)fprintf(fp, "%s class add dev %s parent 1:1 classid 1:2 htb rate %dkbit ceil %dkbit quantum %d > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, cf.rdownlink - cf.pdownlink, cf.rdownlink, cf.cfull_qdisc_opt->htb_quantum);
	
				//p2pwnc bw class for inet_dev - UPLINK
				(void)fprintf(fp, "%s class add dev %s parent 1:1 classid 1:3 htb rate %dkbit ceil %dkbit quantum %d > /dev/null 2>&1\n", cf.tc, cf.inet_dev, cf.puplink, cf.puplink, cf.cfull_qdisc_opt->htb_quantum);
				//p2pwnc bw class for p2pwnc_dev - DOWNLINK
				(void)fprintf(fp, "%s class add dev %s parent 1:1 classid 1:3 htb rate %dkbit ceil %dkbit quantum %d > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, cf.pdownlink, cf.pdownlink, cf.cfull_qdisc_opt->htb_quantum);
			}
			else {
				//full bw class for inet_dev - UPLINK
				(void)fprintf(fp, "%s class add dev %s parent 1:0 classid 1:1 htb rate %dkbit ceil %dkbit > /dev/null 2>&1\n", cf.tc, cf.inet_dev, cf.ruplink, cf.ruplink);
				//full bw class for p2pwnc_dev - DOWNLINK
				(void)fprintf(fp, "%s class add dev %s parent 1:0 classid 1:1 htb rate %dkbit ceil %dkbit > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, cf.rdownlink, cf.rdownlink);

				//home bw class for inet_dev - UPLINK
				(void)fprintf(fp, "%s class add dev %s parent 1:1 classid 1:2 htb rate %dkbit ceil %dkbit > /dev/null 2>&1\n", cf.tc, cf.inet_dev, cf.ruplink - cf.puplink, cf.ruplink);
				//home bw class for p2pwnc_dev - DOWNLINK
				(void)fprintf(fp, "%s class add dev %s parent 1:1 classid 1:2 htb rate %dkbit ceil %dkbit > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, cf.rdownlink - cf.pdownlink, cf.rdownlink);
	
				//p2pwnc bw class for inet_dev - UPLINK
				(void)fprintf(fp, "%s class add dev %s parent 1:1 classid 1:3 htb rate %dkbit ceil %dkbit > /dev/null 2>&1\n", cf.tc, cf.inet_dev, cf.puplink, cf.puplink);
				//p2pwnc bw class for p2pwnc_dev - DOWNLINK
				(void)fprintf(fp, "%s class add dev %s parent 1:1 classid 1:3 htb rate %dkbit ceil %dkbit > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, cf.pdownlink, cf.pdownlink);
			}
			switch(cf.cless_qdisc) {	
				case BFIFO: //use bfifo as qdisc 
				if(cf.cless_qdisc_opt != NULL) {//bfifo with options
					//home qdisc for inet_dev - UPLINK
					(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 bfifo limit %d > /dev/null 2>&1\n", cf.tc, cf.inet_dev, cf.cless_qdisc_opt->fifo_limit);
					//home qdisc for p2pwnc_dev - DOWNLINK
					(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 bfifo limit %d > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, cf.cless_qdisc_opt->fifo_limit);
				}
				else {
					//home qdisc for inet_dev - UPLINK
					(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 bfifo > /dev/null 2>&1\n", cf.tc, cf.inet_dev);
					//home qdisc for p2pwnc_dev - DOWNLINK
					(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 bfifo > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev);
				}
				break;
				case PFIFO: //use pfifo as qdisc 
				if(cf.cless_qdisc_opt != NULL) { //pfifo options
					//home qdisc for inet_dev - UPLINK
					(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 pfifo limit %d > /dev/null 2>&1\n", cf.tc, cf.inet_dev, cf.cless_qdisc_opt->fifo_limit);
					//home qdisc for p2pwnc_dev - DOWNLINK
					(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 pfifo limit %d > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, cf.cless_qdisc_opt->fifo_limit);
				}
				else {
					//home qdisc for inet_dev - UPLINK
					(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 pfifo > /dev/null 2>&1\n", cf.tc, cf.inet_dev);
					//home qdisc for p2pwnc_dev - DOWNLINK
					(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 pfifo > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev);
				}
				break;
				case SFQ: //use sfq as qdisc 
				if(cf.cless_qdisc_opt != NULL) { //sfq options
					if(cf.cless_qdisc_opt->sfq_perturb != 0 && cf.cless_qdisc_opt->sfq_quantum != 0 && cf.cless_qdisc_opt->fifo_limit != 0) {
						//home qdisc for inet_dev - UPLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 sfq perturb %d  quantum %d limit %d > /dev/null 2>&1\n", cf.tc, cf.inet_dev, cf.cless_qdisc_opt->sfq_perturb, cf.cless_qdisc_opt->sfq_quantum, cf.cless_qdisc_opt->fifo_limit);
						//home qdisc for p2pwnc_dev - DOWNLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 sfq perturb %d quantum %d limit %d > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, cf.cless_qdisc_opt->sfq_perturb, cf.cless_qdisc_opt->sfq_quantum, cf.cless_qdisc_opt->fifo_limit);
					}
					else {
						if(cf.cless_qdisc_opt->sfq_perturb != 0) {
							(void)sprintf(cmdbuf, "%s qdisc add dev %s handle 2: parent 1:2 sfq perturb %d", cf.tc, cf.inet_dev, cf.cless_qdisc_opt->sfq_perturb);
							if(cf.cless_qdisc_opt->sfq_quantum != 0) 
								(void)sprintf(cmdbuf, "%s quantum %d", cmdbuf, cf.cless_qdisc_opt->sfq_quantum);
							if(cf.cless_qdisc_opt->fifo_limit != 0) 
								(void)sprintf(cmdbuf, "%s limit %d", cmdbuf, cf.cless_qdisc_opt->fifo_limit);
							(void)fprintf(fp, "%s > /dev/null 2>&1\n", cmdbuf);
							
							(void)sprintf(cmdbuf, "%s qdisc add dev %s handle 2: parent 1:2 sfq perturb %d", cf.tc, cf.p2pwnc_dev, cf.cless_qdisc_opt->sfq_perturb);
							if(cf.cless_qdisc_opt->sfq_quantum != 0) 
								(void)sprintf(cmdbuf, "%s quantum %d", cmdbuf, cf.cless_qdisc_opt->sfq_quantum);
							if(cf.cless_qdisc_opt->fifo_limit != 0) 
								(void)sprintf(cmdbuf, "%s limit %d", cmdbuf, cf.cless_qdisc_opt->fifo_limit);
							(void)fprintf(fp, "%s > /dev/null 2>&1\n", cmdbuf);
							break;
						}	
						if(cf.cless_qdisc_opt->sfq_quantum != 0) {
							(void)sprintf(cmdbuf, "%s qdisc add dev %s handle 2: parent 1:2 sfq quantum %d", cf.tc, cf.inet_dev, cf.cless_qdisc_opt->sfq_quantum);
							if(cf.cless_qdisc_opt->sfq_perturb != 0) 
								(void)sprintf(cmdbuf, "%s perturb %d", cmdbuf, cf.cless_qdisc_opt->sfq_perturb);
							if(cf.cless_qdisc_opt->fifo_limit != 0) 
								(void)sprintf(cmdbuf, "%s limit %d", cmdbuf, cf.cless_qdisc_opt->fifo_limit);
							(void)fprintf(fp, "%s > /dev/null 2>&1\n", cmdbuf);
							
							(void)sprintf(cmdbuf, "%s qdisc add dev %s handle 2: parent 1:2 sfq quantum %d", cf.tc, cf.p2pwnc_dev, cf.cless_qdisc_opt->sfq_quantum);
							if(cf.cless_qdisc_opt->sfq_perturb != 0) 
								(void)sprintf(cmdbuf, "%s perturb %d", cmdbuf, cf.cless_qdisc_opt->sfq_perturb);
							if(cf.cless_qdisc_opt->fifo_limit != 0) 
								(void)sprintf(cmdbuf, "%s limit %d", cmdbuf, cf.cless_qdisc_opt->fifo_limit);
							(void)fprintf(fp, "%s > /dev/null 2>&1\n", cmdbuf);
							break;
						}
						else{
							(void)sprintf(cmdbuf, "%s qdisc add dev %s handle 2: parent 1:2 sfq limit %d", cf.tc, cf.inet_dev, cf.cless_qdisc_opt->fifo_limit);
							if(cf.cless_qdisc_opt->sfq_perturb != 0) 
								(void)sprintf(cmdbuf, "%s perturb %d", cmdbuf, cf.cless_qdisc_opt->sfq_perturb);
							if(cf.cless_qdisc_opt->sfq_quantum != 0) 
								(void)sprintf(cmdbuf, "%s quantum %d", cmdbuf, cf.cless_qdisc_opt->sfq_quantum);
							(void)fprintf(fp, "%s > /dev/null 2>&1\n", cmdbuf);
							
							(void)sprintf(cmdbuf, "%s qdisc add dev %s handle 2: parent 1:2 sfq limit %d", cf.tc, cf.p2pwnc_dev, cf.cless_qdisc_opt->fifo_limit);
							if(cf.cless_qdisc_opt->sfq_perturb != 0) 
								(void)sprintf(cmdbuf, "%s perturb %d", cmdbuf, cf.cless_qdisc_opt->sfq_perturb);
							if(cf.cless_qdisc_opt->sfq_quantum != 0) 
								(void)sprintf(cmdbuf, "%s quantum %d", cmdbuf, cf.cless_qdisc_opt->sfq_quantum);
							(void)fprintf(fp, "%s > /dev/null 2>&1\n", cmdbuf);
							break;
						}
								
					}				
				}
				else {
					//home qdisc for inet_dev - UPLINK
					(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 sfq > /dev/null 2>&1\n", cf.tc, cf.inet_dev);
					//home qdisc for p2pwnc_dev - DOWNLINK
					(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 sfq > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev);
					break;
				}
				case RED: //use red as qdisc - EXPERIMENTAL

					max = min = limit = burst = 0;
					//calculate values for inet_dev - UPLINK
					max = (((cf.ruplink - cf.puplink) * cf.cless_qdisc_opt->red_latency * 1000) / (8000));
					min = max / 3;
					limit = 8 * max;
					burst = (((2 * min) + max) / (3 * cf.cless_qdisc_opt->red_avpkt));
					if(cf.cless_qdisc_opt->red_ecn == 0)
						(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 red limit %d max %d min %d avpkt %d burst %d probability %f.3 bandwidth %d > /dev/null 2>&1\n", cf.tc, cf.inet_dev, limit, max, min, cf.cless_qdisc_opt->red_avpkt, burst, cf.cless_qdisc_opt->red_probability, cf.ruplink - cf.puplink);
					else
						(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 red limit %d max %d min %d avpkt %d burst %d probability %f.3 bandwidth %d ecn > /dev/null 2>&1\n", cf.tc, cf.inet_dev, limit, max, min, cf.cless_qdisc_opt->red_avpkt, burst, cf.cless_qdisc_opt->red_probability, cf.ruplink - cf.puplink);
					
					max = min = limit = burst = 0;
					//calculate values for inet_dev - DOWNLINK
					max = (((cf.rdownlink - cf.pdownlink) * cf.cless_qdisc_opt->red_latency * 1000) / (8000));
					min = max / 3;
					limit = 8 * max;
					burst = (((2 * min) + max) / (3 * cf.cless_qdisc_opt->red_avpkt));
					if(cf.cless_qdisc_opt->red_ecn == 0)
						(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 red limit %d max %d min %d avpkt %d burst %d probability %f.3 bandwidth %d > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, limit, max, min, cf.cless_qdisc_opt->red_avpkt, burst, cf.cless_qdisc_opt->red_probability, cf.rdownlink - cf.pdownlink);
					else
						(void)fprintf(fp, "%s qdisc add dev %s handle 2: parent 1:2 red limit %d max %d min %d avpkt %d burst %d probability %f.3 bandwidth %d ecn > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, limit, max, min, cf.cless_qdisc_opt->red_avpkt, burst, cf.cless_qdisc_opt->red_probability, cf.rdownlink - cf.pdownlink);
				break;
			}	
			//traverse allocation pool to generate the appropriate tc and iptables rules for each client
			clp = NULL;
			clp = rpool.cls;
			i = 4; //first available class
			
			while(clp != NULL) { //for each client
				
				if(cf.cfull_qdisc_opt != NULL && cf.cfull_qdisc_opt->htb_quantum != 0) {
					//client bw class for inet_dev - UPLINK
					(void)fprintf(fp, "%s class add dev %s parent 1:3 classid 1:%d htb rate %dkbit ceil %dkbit quantum %d > /dev/null 2>&1\n", cf.tc, cf.inet_dev, i, (clp->qos_level * cf.puplink)/rpool.qos_level_sum, cf.puplink, cf.cfull_qdisc_opt->htb_quantum);
					//client bw class for p2pwnc_dev - DOWNLINK
					(void)fprintf(fp, "%s class add dev %s parent 1:3 classid 1:%d htb rate %dkbit ceil %dkbit quantum %d > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, i, (clp->qos_level * cf.pdownlink)/rpool.qos_level_sum, cf.pdownlink, cf.cfull_qdisc_opt->htb_quantum);
				}
				else {
					//client bw class for inet_dev - UPLINK
					(void)fprintf(fp, "%s class add dev %s parent 1:3 classid 1:%d htb rate %dkbit ceil %dkbit > /dev/null 2>&1\n", cf.tc, cf.inet_dev, i, (clp->qos_level * cf.puplink)/rpool.qos_level_sum, cf.puplink);
					//client bw class for p2pwnc_dev - DOWNLINK
					(void)fprintf(fp, "%s class add dev %s parent 1:3 classid 1:%d htb rate %dkbit ceil %dkbit > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, i, (clp->qos_level * cf.pdownlink)/rpool.qos_level_sum, cf.pdownlink);
				}
				
				switch(cf.cless_qdisc) {
					case BFIFO: //use bfifo as qdisc 
					if(cf.cless_qdisc_opt != NULL) {//bfifo with options
						//client qdisc for inet_dev - UPLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d bfifo limit %d > /dev/null 2>&1\n", cf.tc, cf.inet_dev, i, i, cf.cless_qdisc_opt->fifo_limit);
						//client qdisc for p2pwnc_dev - DOWNLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d bfifo limit %d > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, i, i, cf.cless_qdisc_opt->fifo_limit);
					}
					else {
						//client qdisc for inet_dev - UPLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d bfifo > /dev/null 2>&1\n", cf.tc, cf.inet_dev, i, i);
						//client qdisc for p2pwnc_dev - DOWNLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d bfifo > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, i, i);
					}
					break;
					case PFIFO: //use pfifo as qdisc 
					if(cf.cless_qdisc_opt != NULL) { //pfifo options
						//client qdisc for inet_dev - UPLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d pfifo limit %d > /dev/null 2>&1\n", cf.tc, cf.inet_dev, i, i, cf.cless_qdisc_opt->fifo_limit);
						//client qdisc for p2pwnc_dev - DOWNLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d pfifo limit %d > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, i, i, cf.cless_qdisc_opt->fifo_limit);
					}
					else {
						//client qdisc for inet_dev - UPLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d pfifo > /dev/null 2>&1\n", cf.tc, cf.inet_dev, i, i);
						//client qdisc for p2pwnc_dev - DOWNLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d pfifo > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, i, i);
					}
					break;
					case SFQ: //use sfq as qdisc 
					if(cf.cless_qdisc_opt != NULL) { //sfq options
						if(cf.cless_qdisc_opt->sfq_perturb != 0 && cf.cless_qdisc_opt->sfq_quantum != 0 && cf.cless_qdisc_opt->fifo_limit != 0) {
						//client qdisc for inet_dev - UPLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d sfq perturb %d  quantum %d limit %d > /dev/null 2>&1\n", cf.tc, cf.inet_dev, i, i, cf.cless_qdisc_opt->sfq_perturb, cf.cless_qdisc_opt->sfq_quantum, cf.cless_qdisc_opt->fifo_limit);
						//client qdisc for p2pwnc_dev - DOWNLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d sfq perturb %d quantum %d limit %d > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, i, i, cf.cless_qdisc_opt->sfq_perturb, cf.cless_qdisc_opt->sfq_quantum, cf.cless_qdisc_opt->fifo_limit);
						}
						else {
						if(cf.cless_qdisc_opt->sfq_perturb != 0) {
							//inet_dev
							(void)memset(cmdbuf, '\0', 2*MSGBUF_MAX);
							(void)sprintf(cmdbuf, "%s qdisc add dev %s handle %d: parent 1:%d sfq perturb %d", cf.tc, cf.inet_dev, i, i, cf.cless_qdisc_opt->sfq_perturb);
							if(cf.cless_qdisc_opt->sfq_quantum != 0) 
								(void)sprintf(cmdbuf, "%s quantum %d", cmdbuf, cf.cless_qdisc_opt->sfq_quantum);
							if(cf.cless_qdisc_opt->fifo_limit != 0) 
								(void)sprintf(cmdbuf, "%s limit %d", cmdbuf, cf.cless_qdisc_opt->fifo_limit);
							(void)fprintf(fp, "%s > /dev/null 2>&1\n", cmdbuf);
							//p2pwnc_dev
							(void)memset(cmdbuf, '\0', 2*MSGBUF_MAX);
							(void)sprintf(cmdbuf, "%s qdisc add dev %s handle %d: parent 1:%d sfq perturb %d", cf.tc, cf.p2pwnc_dev, i, i, cf.cless_qdisc_opt->sfq_perturb);
							if(cf.cless_qdisc_opt->sfq_quantum != 0) 
								(void)sprintf(cmdbuf, "%s quantum %d", cmdbuf, cf.cless_qdisc_opt->sfq_quantum);
							if(cf.cless_qdisc_opt->fifo_limit != 0) 
								(void)sprintf(cmdbuf, "%s limit %d", cmdbuf, cf.cless_qdisc_opt->fifo_limit);
							(void)fprintf(fp, "%s > /dev/null 2>&1\n", cmdbuf);
							break;
							}	
						if(cf.cless_qdisc_opt->sfq_quantum != 0) {
							//inet_dev
							(void)memset(cmdbuf, '\0', 2*MSGBUF_MAX);
							(void)sprintf(cmdbuf, "%s qdisc add dev %s handle %d: parent 1:%d sfq quantum %d", cf.tc, cf.inet_dev, i,i,cf.cless_qdisc_opt->sfq_quantum);
							if(cf.cless_qdisc_opt->sfq_perturb != 0) 
								(void)sprintf(cmdbuf, "%s perturb %d", cmdbuf, cf.cless_qdisc_opt->sfq_perturb);
							if(cf.cless_qdisc_opt->fifo_limit != 0) 
								(void)sprintf(cmdbuf, "%s limit %d", cmdbuf, cf.cless_qdisc_opt->fifo_limit);
							(void)fprintf(fp, "%s > /dev/null 2>&1\n", cmdbuf);
							//p2pwnc_dev
							(void)memset(cmdbuf, '\0', 2*MSGBUF_MAX);
							(void)sprintf(cmdbuf, "%s qdisc add dev %s handle %d: parent 1:%d sfq quantum %d", cf.tc, cf.p2pwnc_dev, i,i, cf.cless_qdisc_opt->sfq_quantum);
							if(cf.cless_qdisc_opt->sfq_perturb != 0) 
								(void)sprintf(cmdbuf, "%s perturb %d", cmdbuf, cf.cless_qdisc_opt->sfq_perturb);
							if(cf.cless_qdisc_opt->fifo_limit != 0) 
								(void)sprintf(cmdbuf, "%s limit %d", cmdbuf, cf.cless_qdisc_opt->fifo_limit);
							(void)fprintf(fp, "%s > /dev/null 2>&1\n", cmdbuf);
							break;
						}
						else{
							//inet_dev
							(void)memset(cmdbuf, '\0', 2*MSGBUF_MAX);
							(void)sprintf(cmdbuf, "%s qdisc add dev %s handle %d: parent 1:%d sfq limit %d", cf.tc, cf.inet_dev, i, i, cf.cless_qdisc_opt->fifo_limit);
							if(cf.cless_qdisc_opt->sfq_perturb != 0) 
								(void)sprintf(cmdbuf, "%s perturb %d", cmdbuf, cf.cless_qdisc_opt->sfq_perturb);
							if(cf.cless_qdisc_opt->sfq_quantum != 0) 
								(void)sprintf(cmdbuf, "%s quantum %d", cmdbuf, cf.cless_qdisc_opt->sfq_quantum);
							(void)fprintf(fp, "%s > /dev/null 2>&1\n", cmdbuf);
							//p2pwnc_dev
							(void)memset(cmdbuf, '\0', 2*MSGBUF_MAX);
							(void)sprintf(cmdbuf, "%s qdisc add dev %s handle %d: parent 1:%d sfq limit %d", cf.tc, cf.p2pwnc_dev, i,i, cf.cless_qdisc_opt->fifo_limit);
							if(cf.cless_qdisc_opt->sfq_perturb != 0) 
								(void)sprintf(cmdbuf, "%s perturb %d", cmdbuf, cf.cless_qdisc_opt->sfq_perturb);
							if(cf.cless_qdisc_opt->sfq_quantum != 0) 
								(void)sprintf(cmdbuf, "%s quantum %d", cmdbuf, cf.cless_qdisc_opt->sfq_quantum);
							(void)fprintf(fp, "%s > /dev/null 2>&1\n", cmdbuf);
							break;
							}
								
						}				
					}
					else {
						//client qdisc for inet_dev - UPLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d sfq > /dev/null 2>&1\n", cf.tc, cf.inet_dev, i,i);
						//client qdisc for p2pwnc_dev - DOWNLINK
						(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d sfq > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, i,i);
						break;
					}
					case RED: //use red as qdisc - EXPERIMENTAL
						max = min = limit = burst = 0;
						//calculate values for inet_dev - UPLINK
						max = ((((clp->qos_level * cf.puplink) / rpool.qos_level_sum) * cf.cless_qdisc_opt->red_latency * 1000) / (8000));
						min = max / 3;
						limit = 8 * max;
						burst = (((2 * min) + max) / (3 * cf.cless_qdisc_opt->red_avpkt));
						if(cf.cless_qdisc_opt->red_ecn == 0)
							(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d red limit %d max %d min %d avpkt %d burst %d probability %f.3 bandwidth %d > /dev/null 2>&1\n", cf.tc, cf.inet_dev, i,i,limit, max, min, cf.cless_qdisc_opt->red_avpkt, burst, cf.cless_qdisc_opt->red_probability, ((clp->qos_level * cf.puplink) / rpool.qos_level_sum));
						else
							(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d red limit %d max %d min %d avpkt %d burst %d probability %f.3 bandwidth %d ecn > /dev/null 2>&1\n", cf.tc, cf.inet_dev, i,i,limit, max, min, cf.cless_qdisc_opt->red_avpkt, burst, cf.cless_qdisc_opt->red_probability, ((clp->qos_level * cf.puplink) / rpool.qos_level_sum));
						max = min = limit = burst = 0;
						//calculate values for inet_dev - DOWNLINK
						max = ((((clp->qos_level * cf.pdownlink) / rpool.qos_level_sum) * cf.cless_qdisc_opt->red_latency * 1000) / (8000));
						min = max / 3;
						limit = 8 * max;
						burst = (((2 * min) + max) / (3 * cf.cless_qdisc_opt->red_avpkt));
						if(cf.cless_qdisc_opt->red_ecn == 0)
							(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d red limit %d max %d min %d avpkt %d burst %d probability %f.3 bandwidth %d > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, i,i, limit, max, min, cf.cless_qdisc_opt->red_avpkt, burst, cf.cless_qdisc_opt->red_probability, ((clp->qos_level * cf.pdownlink) / rpool.qos_level_sum));
						else
							(void)fprintf(fp, "%s qdisc add dev %s handle %d: parent 1:%d red limit %d max %d min %d avpkt %d burst %d probability %f.3 bandwidth %d ecn > /dev/null 2>&1\n", cf.tc, cf.p2pwnc_dev, i,i,limit, max, min, cf.cless_qdisc_opt->red_avpkt, burst, cf.cless_qdisc_opt->red_probability, ((clp->qos_level * cf.pdownlink) / rpool.qos_level_sum));
					break;
				}	
				//update filters
				//rule for inet_dev - UPLINK
				(void)fprintf(fp, "%s -t mangle -A PREROUTING -i %s -s %s -j MARK --set-mark %d > /dev/null 2>&1\n", cf.iptables, cf.p2pwnc_dev, inet_ntoa(clp->ip_address) ,i);
				//rule for p2pwnc_dev - DOWNLINK
				(void)fprintf(fp, "%s -t mangle -A POSTROUTING -o %s -d %s -j MARK --set-mark %d > /dev/null 2>&1\n", cf.iptables, cf.p2pwnc_dev, inet_ntoa(clp->ip_address) ,i);
				//filter for inet_dev - UPLINK
				(void)fprintf(fp, "%s filter add dev %s protocol ip parent 1:0 prio 1 handle %d fw flowid 1:%d > /dev/null 2>&1\n",cf.tc, cf.inet_dev, i, i);
				//filter for p2pwnc_dev - DOWNLINK
				(void)fprintf(fp, "%s filter add dev %s protocol ip parent 1:0 prio 1 handle %d fw flowid 1:%d > /dev/null 2>&1\n",cf.tc, cf.p2pwnc_dev, i, i);
				i++;//increase the first available class
				clp = (p2pwnc_client*)clp->next;//get next client

			}
			break;
		}
		//remove tc script - exec
		(void)fclose(fp);
		(void)memset(cmdbuf, 0, sizeof(cmdbuf));
		(void)sprintf(cmdbuf,"/bin/sh %s",TMPFILE);
		if(system(cmdbuf) == -1) {
			(void)unlink(TMPFILE);  
			return 0;
		}
		else {
			(void)unlink(TMPFILE);  
			return 1;
		}
	}
}

//################################################################

//################################################################
/* 
 * Check if msg is a valid protocol type header - Only URRQ message
 * is accepted. Return 1 for true or 0 for false
 */
uint8_t is_valid_header(char* msg) {
	if(strlen(msg) != 17) //invalid length
		return 0;
	if(!strncmp(msg, "URRQ P2PWNC/", 12)) { //validate the first 12 bytes
		if(!strncmp(msg+15, "\r\n", 2)) { //validate the last two bytes CRLF
			if(msg[13] == '.') {//validate the protocol number
				if(isdigit(*(msg+12))) {
					if(isdigit(*(msg+14)))
						return 1; // YES!!!
				}
			}
		}
	} 
	return 0;
}
//################################################################

//################################################################
/* 
 * Check if msg is a valid protocol URRQ content message.
 * Return: 1 for true and store message ip address, qos level 
 * and team member values in the coresponding variables
 * Return: 0 for false
 */
uint8_t is_valid_content(
			uint8_t length, //message length 
			char* msg, //message buffer
			struct in_addr* address, //message content address
			uint8_t* team, //message content team member
			uint8_t* qlevel) { //message content qos level
	
	int i, index; 
		
	//validate
	if(strlen(msg) != length) //corrupted message
		return 0;

	for(i=0; i<strlen(msg); i++) //convert to lower case
		msg[i] = tolower(msg[i]);

	if(strncmp(msg+length-2, "\r\n", 2)) //missing final CRLF
		return 0;
	msg[length-2] = '\0'; //ignore last CRLF (validated above)
	
	switch(msg[length-4]) {
		//one byte qos level value
		case ' ':
			//check for valid byte
			if(isdigit(msg[length-3])) {
				//get qos level value
				*qlevel = msg[length-3] - 48;
				msg[length-4] = '\0';
				break;
			}
			//error
			return 0;
		//two or more bytes qos level
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			break;
		//invalid header
		default:
			return 0;
	}
	
	//two or three bytes qos level value
	if((isdigit(msg[length-5]) || msg[length-5] == ' ') && msg[length-4] != '\0') {
		if(isdigit(msg[length-3])) {
			
			if(atoi(msg+length-5) > 0 && atoi(msg+length-5) < 256) {
				//get qos level value
				*qlevel = atoi(msg+length-5);
				if(msg[length-5] == ' ')
					msg[length-5] = '\0';
				else
					msg[length-6] = '\0';
			}
			else
				return 0;
		}
		else //error
			return 0;
	}
	else//error
		if(!(msg[length-4] == '\0'))
			return 0;
	
	//validate
	if(strncmp(msg+strlen(msg)-28, "\r\nteam-member: y\r\nqos-level:", 28) && strncmp(msg+strlen(msg)-28, "\r\nteam-member: n\r\nqos-level:", 28))	
		return 0;
	
	//get team member value
	if(msg[strlen(msg)-13] == 'y')
		*team = 1;
	else
		*team = 0;
	
	msg[strlen(msg)-28] = '\0';

	//validate ip address
	if(strncmp(msg, "ip-address: ", 12))
		return 0;
	if(!inet_aton(msg+12,address))
		return 0;

	//finaly return 1 - yioupi !-!-!-!
	return 1;
}
//################################################################

//################################################################
/*
 * TC server - service module
 */
void tc_server(int fd) {
	struct sockaddr_in client; //info about client
	size_t client_size = sizeof(client); //client struct size
	struct sigaction alrm; //alarm sigaction
	char msgbuf[MSGBUF_MAX]; //message buffer
	int i, found; //generic usage
	
	uint8_t clen; //content length 
	uint8_t team_member; //local = 1 or p2pwnc = 0 (default)
	uint8_t qlevel; //qos level [0,255] 
	struct in_addr address; //message ip address
	
	//syslog info message for server init
	(void)memset(msgbuf, '\0', MSGBUF_MAX);
	(void)sprintf(msgbuf, "Waiting for new connections on port: %d", ntohs(cf.port));
	syslog(LOG_FACILITY | LOG_INFO, msgbuf);
	
	//init resource pool
	(void)memset(&rpool, 0, sizeof(resource_pool));
	rpool.num = 0; //zero clients for startup
	rpool.qos_level_sum = 0; //zero sum for startup
	rpool.cls = NULL;

	//accept a new connection 
	while(cfd = accept(fd, (struct sockaddr *) &client, (size_t *)&client_size)) { //iterative server 
			
		if(cfd == -1) {//error
			fail("while trying to process an incomming connection. Ignoring\0", 6);
			continue;//get a new connection 
		}
		
		//process the connection
		//syslog - client info
		(void)memset(msgbuf, '\0', MSGBUF_MAX);
		(void)sprintf(msgbuf, "Connection from: %s:%d", inet_ntoa(client.sin_addr), ntohs(client.sin_port));
		syslog(LOG_FACILITY | LOG_INFO, msgbuf);
		
		//init
		i = found = 0;
		clen = 0;
		team_member = 0;
		qlevel = 0;
		address.s_addr = 0;
		
		if(bsearch(&client.sin_addr.s_addr, cf.servers, cf.serv_num, sizeof(struct in_addr), in_addr_cmp) == NULL){ //unauthorized address
		
			//syslog info
			(void)memset(msgbuf, '\0', MSGBUF_MAX);
			(void)sprintf(msgbuf, "Closing connection. Address: %s is not listed in p2pwnc server list", inet_ntoa(client.sin_addr));
			(void)close(cfd); //close the connection
			fail(msgbuf, 6);
			continue; //get a new connection
		}
		
		//set alarm timeout
		set_timer(&alrm, ACTIVATE_TIMER); // 0->alarm on
		(void)memset(msgbuf, 0, MSGBUF_MAX); //init message buffer
		
		if(recv(cfd, msgbuf, 17, MSG_WAITALL) <= 0) {//GET p2pwnc protocol message type header (17 bytes, 4 bytes type 1 byte space, 10 bytes protocol, 2 bytes CRLF)
			
			if(alarm_sig == 0) //no timeout occured - client aborted
				fail("Closing connection - Client aborted", 6);
			set_timer(&alrm, DEACTIVATE_TIMER); // 1->alarm off
			(void)close(cfd); //close the socket
			continue; //get a new connection
		}
		
		//validate protocol
		if(!is_valid_header(msgbuf)) { //invalid message type header (only URRQ message is accepted)
			
			if(alarm_sig == 0) //no timeout occured before
				fail("Closing connection - Invalid message type protocol header", 6);
			set_timer(&alrm, DEACTIVATE_TIMER); // 1->alarm off
			(void)close(cfd); //close the socket
			continue; //get a new connection
		}
		
		(void)memset(msgbuf, 0, MSGBUF_MAX); //clean message buffer
		if(recv(cfd, msgbuf, 16, MSG_WAITALL) <= 0) {//GET p2pwnc protocol content length header (16 bytes, 15 bytes fixed string - "Content-length:", 1 byte space)
			
			if(alarm_sig == 0) //no timeout occured - client aborted
				fail("Closing connection - Client aborted", 6);
			set_timer(&alrm, DEACTIVATE_TIMER); // 1->alarm off
			(void)close(cfd); //close the socket
			continue; //get a new connection
		}
		//validate content length - inline no call to external function 
		for(i = 0; i < 16; i++) //convert to lower case
			msgbuf[i] = tolower(msgbuf[i]);
		if(strncmp(msgbuf, "content-length: ", 16)) { //invalid content length header
			if(alarm_sig == 0) //no timeout occured before
				fail("Closing connection - Invalid content-length protocol header", 6);
			set_timer(&alrm, DEACTIVATE_TIMER); // 1->alarm off
			(void)close(cfd); //close the socket
			continue; //get a new connection
		}
		(void)memset(msgbuf, 0, MSGBUF_MAX); //clean message buffer
		if(recv(cfd, msgbuf, 4, MSG_WAITALL) <= 0) {//GET p2pwnc protocol content length value (4 bytes, 2 bytes content length value, 2 bytes CRLF)
			
			if(alarm_sig == 0) //no timeout occured - client aborted
				fail("Closing connection - Client aborted", 6);
			set_timer(&alrm, DEACTIVATE_TIMER); // 1->alarm off
			(void)close(cfd); //close the socket
			continue; //get a new connection
		}
		if((strlen(msgbuf) < 4) || (atoi(msgbuf) < 51) || (atoi(msgbuf) > 61 ) || (strncmp(msgbuf+2, "\r\n",2))) { //invalid content length value
			if(alarm_sig == 0)
				fail("Closing connection - Invalid content-length protocol value", 6);
			set_timer(&alrm, DEACTIVATE_TIMER); // 1->alarm off
			(void)close(cfd); //close the socket
			continue; //get a new connection
		}
		
		clen = atoi(msgbuf); //get the valid content length value
		(void)memset(msgbuf, 0, MSGBUF_MAX); //clean message buffer
		if(recv(cfd, msgbuf, clen, MSG_WAITALL) <= 0) {//GET p2pwnc protocol URRQ message content (clen bytes)
			
			if(alarm_sig == 0) //no timeout occured - client aborted
				fail("Closing connection - Client aborted", 6);
			set_timer(&alrm, DEACTIVATE_TIMER); // 1->alarm off
			(void)close(cfd); //close the socket
			continue; //get a new connection
		}
		if(!is_valid_content(clen, msgbuf, &address, &team_member, &qlevel)) {
				
			if(alarm_sig == 0)
				fail("Closing connection - Invalid URRQ message content", 6);
			set_timer(&alrm, DEACTIVATE_TIMER); // 1->alarm off
			(void)close(cfd); //close the socket
			continue; //get a new connection
		}		
		
		if(!update_resource_pool(address, team_member, qlevel)) {//update resource pool
		
			fail("Closing connection - Query for removing a non-existent ip address", 6);
			set_timer(&alrm, DEACTIVATE_TIMER); // 1->alarm off
			(void)close(cfd); //close the socket
			continue; //get a new connection
		}
		
		if(!init_tc(CLEAR_INIT, &address)) //update reservations 
			fail("Failed while trying to update classfull and classless qdiscs", 5);
		else {
			//send URRA - reply answer
			if(send(cfd, "URRA P2PWNC/3.0\r\n", 17, 0) == -1) 
				fail("Closing connection - Failed to send URRA reply answer", 6);
		}
			
		//remove alarm handler
		set_timer(&alrm, DEACTIVATE_TIMER); // 1->alarm off
		(void)close(cfd); //cleanup
		//inform syslog
		log_updated_tc();
 	}
}
//################################################################

//################################################################
/* 
 * Error handler. This function is called after each error. 
 * It displays (syslog) a valid error message and and terminates 
 * program's execution if the error is fatal. 
*/ 
void fail(char* reason, int retvalue) {
	
	switch(retvalue) {
	//Normal termination - for daemon process (close sysloging)
	case -1: 
 		syslog(LOG_FACILITY | LOG_INFO, "Stoping P2PWNC Traffic Controller");
		closelog();
		exit(0); //normal termination
	//Normal terminaton - for not daemon process
	case 0: 
		exit(0); //normal termination
	//Fatal errors - Reported to syslog as CRITICAL 
	//Memory allocation errors
	//Invalid configuration file
	//Failed to open configuration file
	//Error in mandatory entries (configuration file)
	//Missing/invalid program arguments 
	//Socket initialization/bind/listen errors
	case 1: 
	case 2: 
	case 3: 
	case 7: 
 		syslog(LOG_FACILITY | LOG_CRIT, "Error: %s", reason);
 		syslog(LOG_FACILITY | LOG_INFO, "Stoping P2PWNC Traffic Controller");
		closelog();
		exit(retvalue);
	//Non fatal errors - Reported to syslog as WARNING
	//Invalid option in configuration file
	//Error in closing file/network streams
	//Errors while updating tc/iptables
	case 4: 
	case 5: 
 		syslog(LOG_FACILITY | LOG_WARNING, "Warning: %s", reason);
		break;
	//Non fatal errors - Reported to syslog as NOTICE
	//Malformed P2PWNC Protocol headers
	//Unauthorized clients
	//Alarms
	//Connection errors
	case 6:
 		syslog(LOG_FACILITY | LOG_NOTICE, "Notice: %s", reason);
		break;
	}	
}
//################################################################

//################################################################
/* Shutdown signal handler
 * Called for signals: 
 * SIGHUP,
 * SIGINT,
 * SIGQUIT,
 * SIGABRT,
 * SIGTERM
 */
void terminate(int sig) {
	//clear tc rules
	(void)init_tc(CLEAR, NULL);

	(void)close(cfd); //close open connection
	(void)unlink(pidfilename); //remove pid filename
	
	fail(NULL, -1); //terminate
}

//################################################################

//################################################################
/* 
 * Alarm handler - Close the active connection 
 */
void close_connection(int sig) {
	//info message
	syslog(LOG_FACILITY | LOG_NOTICE, "Clossing connection - Alarm! %d secs passed", cf.timeout);
	//set alarm event flag on
	alarm_sig = 1;
}
//################################################################

//################################################################
/* 
 * Parse the configuration file and fill config_file stracture 
 */
void read_config(char* path) {
	//use global variables
	extern config_file cf;
	
	//config file - parsing variables
	FILE* fp; //file pointer
	char *line = NULL; //line buffer
	size_t size = 0; //buffer size
	ssize_t ret; //bytes read
	int i; //generic counter
	
	// Open input stream 
	if(strncmp(path, "-", 1) == 0) //check if we are reading from a file or stdin
		fp = freopen(NULL, "r", stdin); //open stdin for reading
	else
		fp = fopen(path, "r"); //open config file for reading 
	// Check if input stream is open 
	if(fp == NULL)
		fail("while opening configuration file\0",3);
		
	// Fetch lines 	
	while((ret = getline(&line, &size, fp)) != -1) {
	
		char *token; //token buffer
		char *delim = " \t\n"; //delimiters (whitechars and \n)	    
		
		//Tolower case
		for(i = 0; i<size; i++)
			line[i] = tolower(line[i]);
		
		// Tokenize the line 
		token = strtok(line, delim);
		if( token != NULL) {//fetch the first token
			if(!strncmp(token, "inet-dev", 8)) {//process inet dev
				token = strtok(NULL, delim); //try to get network device
				if(token == NULL) //missing net device
					fail("missing inet device in configuration file\0", 2);
				
				//validate the network device
				struct if_nameindex* allifaces; //network interface list
				int found = 0; //generic boolean for searching (0 = false , 0 != true)
				i = 0;
				
				allifaces = if_nameindex(); //get all available system network interfaces
				while((allifaces+i)->if_name != NULL) {
					if(!strncmp((allifaces+i)->if_name, token, strlen((allifaces+i)->if_name))) {
						found = 1; // the interface is ok
						break;
					}
					i++;
				}
				if(!found)
					fail("invalid inet device argument in configuration file\0", 2);
				if_freenameindex(allifaces);
				//update configuration file struct
				cf.inet_dev = (char *)malloc(strlen(token)+1);
				if(cf.inet_dev == NULL)
					fail(strerror(errno),1);
				(void)memset(cf.inet_dev, '\0', strlen(token)+1);
				(void) strncpy(cf.inet_dev,token,strlen(token));
				token = strtok(NULL, delim);//check if there any more tokens left
				if(token != NULL) {
					if(token[0] != '#')//check for comments
						fail("garbage after inet device argument in configuration file\0", 5);
				}
				continue;
			}
			else if(!strncmp(token, "p2pwnc-dev", 10)) {//process p2pwnc dev
				token = strtok(NULL, delim); //try to get network device
				if(token == NULL) //missing net device
					fail("missing p2pwnc device argument in configuration file\0", 2);
				
				//validate the network device
				struct if_nameindex* allifaces; //network interface list
				int found = 0; //generic boolean for searching (0 = false , 0 != true)
				i = 0;
				
				allifaces = if_nameindex(); //get all available system network interfaces
				while((allifaces+i)->if_name != NULL) {
					if(!strncmp((allifaces+i)->if_name, token, strlen((allifaces+i)->if_name))) {
						found = 1; // the interface is ok
						break;
					}
					i++;
				}
				if(!found)
					fail("invalid p2pwnc device argument in configuration file\0", 2);
				if_freenameindex(allifaces);
				//update configuration file struct
				cf.p2pwnc_dev = (char *)malloc(strlen(token)+1);
				if(cf.p2pwnc_dev == NULL)
					fail(strerror(errno),1);
				(void)memset(cf.p2pwnc_dev, '\0', strlen(token)+1);
				(void) strncpy(cf.p2pwnc_dev,token,strlen(token));
				token = strtok(NULL, delim);//check if there any more tokens left
				if(token != NULL) {
					if(token[0] != '#')//check for comments
						fail("garbage after p2pwnc device argument in configuration file\0", 5);
				}
				continue;
			}
			else if(!strncmp(token, "real-uplink", 11)) {//real uplink speed
				token = strtok(NULL, delim); //try to get uplink speed
				if(token == NULL) //missing speed number
					fail("missing real uplink argument in configuration file\0", 2);
				if(!atoi(token)) //real uplink argument is invalid
					fail("invalid real uplink argument in configuration file\0", 2);
				//update configuration file struct
				cf.ruplink = atoi(token);
				token = strtok(NULL, delim);//check if there any more tokens left
				if(token != NULL) {
					if(token[0] != '#')//check for comments
						fail("garbage after real uplink argument in configuration file\0", 5);
				}
				continue;
			}
			else if(!strncmp(token, "real-downlink", 13)) {//real downlink speed
				token = strtok(NULL, delim); //try to get downlink speed
				if(token == NULL) //missing speed number
					fail("missing real downlink argument in configuration file\0", 2);
				if(!atoi(token)) //real downlink argument is invalid
					fail("invalid real downlink argument in configuration file\0", 2);
				//update configuration file struct
				cf.rdownlink = atoi(token);
				token = strtok(NULL, delim);//check if there are any more tokens left
				if(token != NULL) {
					if(token[0] != '#') //check for comments	
						fail("garbage after real downlink argument in configuration file\0", 5);
				}
				continue;
			}
			else if(!strncmp(token, "p2pwnc-uplink", 13)) {//process p2pwnc uplink speed
				token = strtok(NULL, delim); //try to get p2pwnc uplink speed
				if(token == NULL) {//missing speed number
					fail("missing p2pwnc uplink argument in configuration file. Using 50% (default)\0", 5);
					continue;
				}
				if(!atoi(token)) {//p2pwnc uplink argument is invalid
					fail("invalid p2pwnc uplink argument in configuration file. Using 50% (default)\0", 5);
					continue;
				}
				//update configuration file struct
				cf.puplink = atoi(token);
				token = strtok(NULL, delim);//check if there are any more tokens left
				if(token != NULL) {
					if(token[0] != '#') 
						fail("garbage after p2pwnc uplink argument in configuration file\0", 5);
				}
				continue;
			}
			else if(!strncmp(token, "p2pwnc-downlink", 15)) {//process p2pwnc downlink speed
				token = strtok(NULL, delim); //try to get p2pwnc downlink speed
				if(token == NULL) {//missing speed number
					fail("missing p2pwnc downlink argument in configuration file. Using 50% (default)\0", 5);
					continue;
				}
				if(!atoi(token)) {//p2pwnc downlink argument is invalid
					fail("invalid p2pwnc downlink argument in configuration file. Using 50% (default)\0", 5);
					continue;
				}
				//update configuration file struct
				cf.pdownlink = atoi(token);
				token = strtok(NULL, delim);//check if there are any more tokens left
				if(token != NULL) {
					if(token[0] != '#') 
						fail("garbage after p2pwnc downlink argument in configuration file\0", 5);
				}
				continue;
			}
			else if(!strncmp(token, "port", 4)) {//process port
				token = strtok(NULL, delim); //try to get port number
				if(token == NULL) {//missing port number
					fail("missing port number argument in configuration file. Using 7777 (default)\0", 5);
					continue;
				}
				if(!atoi(token)) { //port number argument is invalid 
					fail("invalid port number argument in configuration file. Using 7777 (default)\0", 5);
					continue;
				}
				//update configuration file struct
				cf.port = htons(atoi(token)); //stored in network format (big-endian)
				token = strtok(NULL, delim);//check if there are any more tokens left
				if(token != NULL) {
					if(token[0] != '#')
						fail("garbage after port number argument in configuration file\0", 5);
				}
				continue;
			}
			else if(!strncmp(token, "tc-path", 7)) {//process tc-path
				token = strtok(NULL, delim); //try to get the path
				if(token == NULL) {//missing path
					fail("missing tc path argument in configuration file. Using /sbin/tc (default)\0", 5);
					continue;
				}
				//validate tc path and check for exec with real UID/GID
				if(access(token, X_OK|F_OK) == -1) { //file not exist or file is not exec with real UID/GID
					#ifdef NO_LINKSYS
					if(euidaccess(token, X_OK|F_OK) == -1) { //file not exist or file is not exec with efective UID/GID
					#endif
						fail("invalid tc executable or insufficient access. Using /sbin/tc (default)\0", 5);
						continue;
					#ifdef NO_LINKSYS
					}
					#endif
				}
				//update configuration file struct
				cf.tc = (char *)malloc(strlen(token)+1);
				if(cf.tc == NULL)
					fail(strerror(errno),1);
				(void)memset(cf.tc, '\0', strlen(token)+1);
				(void) strncpy(cf.tc,token,strlen(token));
				token = strtok(NULL, delim);//check if there are any more tokens left
				if(token != NULL) {
					if(token[0] != '#') 
						fail("garbage after tc path argument in configuration file\0", 5);
				}
				continue;
				
			}
			else if(!strncmp(token, "iptables-path", 13)) {//process iptables-path
				token = strtok(NULL, delim); //try to get the path
				if(token == NULL) {//missing path
					fail("missing iptables path argument in configuration file\0", 5);
					continue;
				}
				//validate tc path and check for exec with real UID/GID
				if(access(token, X_OK|F_OK) == -1) { //file not exist or file is not exec with real UID/GID
					#ifdef NO_LINKSYS		
					if(euidaccess(token, X_OK|F_OK) == -1) { //file not exist or file is not exec with efective UID/GID
					#endif
						fail("invalid iptables executable or insufficient access. Using /sbin/iptables (default)\0", 5);
						continue;
					#ifdef NO_LINKSYS
					}
					#endif
				}
				//update configuration file struct
				cf.iptables = (char *)malloc(strlen(token)+1);
				if(cf.iptables == NULL)
					fail(strerror(errno),1);
				(void)memset(cf.iptables, '\0', strlen(token)+1);
				(void) strncpy(cf.iptables,token,strlen(token));
				token = strtok(NULL, delim);//check if there are any more tokens left
				if(token != NULL) {
					if(token[0] != '#')
						fail("garbage after iptables path argument in configuration file\0", 5);
				}
				continue;
			}
			else if(!strncmp(token, "cfull-qdisc", 11)) {//process classfull qdisc
				token = strtok(NULL, delim); //try to get the qdisc
				if(token == NULL) {//missing qdisc
					fail("missing classfull qdisc argument in configuration file. Using htb (default)\0", 5);
					continue;
				}
				//update configuration file struct
				if(!strncmp(token, "htb", 3)) 
					cf.cfull_qdisc = HTB;
				else {
					fail("invalid classfull qdisc argument in configuration file. Using htb (default)\0", 5);
					continue;
				}
				token = strtok(NULL, delim);//check if there are any more tokens left - cfull_qdisc options
				while(token != NULL) {
					if(token[0] == '#') //comments
						break;
					else if(!strncmp(token, "r2q", 3)) { //r2q 
						token = strtok(NULL, delim); //get r2q value
						if(token == NULL) { //missing argument
							fail("missing r2q argument number for classfull qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						//r2q value exists
						if(!atoi(token)) { //validate 
							fail("invalid r2q argument for classfull qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						if(cf.cfull_qdisc_opt == NULL) { //init cfull_qdisc_opt only if it is not initialized
							cf.cfull_qdisc_opt = (cfull_qdisc_options *)malloc(sizeof(cfull_qdisc_options));
							if(cf.cfull_qdisc_opt == NULL)
								fail(strerror(errno),1);
							(void)memset(cf.cfull_qdisc_opt, 0, sizeof(cfull_qdisc_options));
						}
						cf.cfull_qdisc_opt->htb_r2q = atoi(token);
					}
					else if(!strncmp(token, "quantum", 7)) { //quantum 
						token = strtok(NULL, delim); //get quantum value
						if(token == NULL) { //missing argument
							fail("missing quantum argument number for classfull qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						//quantum value exists
						if(!atoi(token)) { //validate 
							fail("invalid quantum argument for classfull qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						if(cf.cfull_qdisc_opt == NULL) {
							cf.cfull_qdisc_opt = (cfull_qdisc_options *)malloc(sizeof(cfull_qdisc_options));
							if(cf.cfull_qdisc_opt == NULL)
								fail(strerror(errno),1);
							(void)memset(cf.cfull_qdisc_opt, 0, sizeof(cfull_qdisc_options));
						}
						cf.cfull_qdisc_opt->htb_quantum = atoi(token);
					}
					else {
						fail("garbage after class full qdisc argument in configuration file\0", 5);
						break;
					}
					token = strtok(NULL, delim); //get next token
				}
				continue;
				
			}
			else if(!strncmp(token, "cless-qdisc", 11)) {//process classless qdisc
				token = strtok(NULL, delim); //try to get the qdisc
				if(token == NULL) {//missing qdisc
					fail("missing classless qdisc argument in configuration file. Using bfifo (default)\0", 5);
					continue;
				}
				//update configuration file struct
				if(!strncmp(token, "bfifo", 5)) 
					cf.cless_qdisc = BFIFO;
				else if(!strncmp(token, "pfifo", 5))
					cf.cless_qdisc = PFIFO;
				else if(!strncmp(token, "sfq", 3))
					cf.cless_qdisc = SFQ;
				else if(!strncmp(token, "red", 3)) {
					cf.cless_qdisc = RED;
					fail("red qdisc is experimental", 5);
				}
				else {
					fail("invalid classless qdisc argument in configuration file. Using bfifo (default)\0", 5);
					continue;
				}
				token = strtok(NULL, delim);//check if there are any more tokens left - cless_qdisc options
				while(token != NULL) {
					
					if(token[0] == '#') //comments
						break;
					else if(!strncmp(token, "limit", 5)) { //limit 
						token = strtok(NULL, delim); //get limit value
						if(token == NULL) { //missing argument
							fail("missing limit argument number for classless qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						//limit value exists
						if(!atoi(token)) { //validate 
							fail("invalid limit argument for classless qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						if(cf.cless_qdisc_opt == NULL) { //init cless_qdisc_opt only if it is not initialized
							cf.cless_qdisc_opt = (cless_qdisc_options *)malloc(sizeof(cless_qdisc_options));
							if(cf.cless_qdisc_opt == NULL)
								fail(strerror(errno),1);
							(void)memset(cf.cless_qdisc_opt, 0, sizeof(cless_qdisc_options));
						}
						cf.cless_qdisc_opt->fifo_limit = atoi(token);
					}
					else if(!strncmp(token, "perturb", 7)) { //perturb 
						token = strtok(NULL, delim); //get perturb value
						if(token == NULL) { //missing argument
							fail("missing perturb argument number for classless qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						//perturb value exists
						if(!atoi(token)) { //validate 
							fail("invalid perturb argument for classless qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						if(cf.cless_qdisc_opt == NULL) {
						
							cf.cless_qdisc_opt = (cless_qdisc_options *)malloc(sizeof(cless_qdisc_options));
							if(cf.cless_qdisc_opt == NULL)
								fail(strerror(errno),1);
							(void)memset(cf.cless_qdisc_opt, 0, sizeof(cless_qdisc_options));
						}
						cf.cless_qdisc_opt->sfq_perturb = atoi(token);
					}
					else if(!strncmp(token, "quantum", 7)) { //quantum 
						token = strtok(NULL, delim); //get quantum value
						if(token == NULL) { //missing argument
							fail("missing quantum argument number for classless qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						//quantum value exists
						if(!atoi(token)) { //validate 
							fail("invalid quantum argument for classless qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						if(cf.cless_qdisc_opt == NULL) {
							cf.cless_qdisc_opt = (cless_qdisc_options *)malloc(sizeof(cless_qdisc_options));
							if(cf.cless_qdisc_opt == NULL)
								fail(strerror(errno),1);
							(void)memset(cf.cless_qdisc_opt, 0, sizeof(cless_qdisc_options));
						}
						cf.cless_qdisc_opt->sfq_quantum = atoi(token);
					}
					else if(!strncmp(token, "latency", 7)) { //latency 
						token = strtok(NULL, delim); //get latency value
						if(token == NULL) { //missing argument
							fail("missing latency argument number for classless qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						//latency value exists
						if(!atoi(token)) { //validate 
							fail("invalid latency argument for classless qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						if(cf.cless_qdisc_opt == NULL) {
							cf.cless_qdisc_opt = (cless_qdisc_options *)malloc(sizeof(cless_qdisc_options));
							if(cf.cless_qdisc_opt == NULL)
								fail(strerror(errno),1);
							(void)memset(cf.cless_qdisc_opt, 0, sizeof(cless_qdisc_options));
						}
						cf.cless_qdisc_opt->red_latency = (size_t)atoi(token);
					}
					else if(!strncmp(token, "probability", 11)) { //probability 
						token = strtok(NULL, delim); //get probability value
						if(token == NULL) { //missing argument
							fail("missing probability argument number for classless qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						//probability value exists
						if(!atof(token)) { //validate 
							fail("invalid probability argument for classless qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						if(cf.cless_qdisc_opt == NULL) {
							cf.cless_qdisc_opt = (cless_qdisc_options *)malloc(sizeof(cless_qdisc_options));
							if(cf.cless_qdisc_opt == NULL)
								fail(strerror(errno),1);
							(void)memset(cf.cless_qdisc_opt, 0, sizeof(cless_qdisc_options));
						}
						cf.cless_qdisc_opt->red_probability = (float)atof(token);
					}
					else if(!strncmp(token, "avpkt", 5)) { //avpkt 
						token = strtok(NULL, delim); //get avpkt value
						if(token == NULL) { //missing argument
							fail("missing avpkt argument number for classless qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						//avpkt value exists
						if(!atoi(token)) { //validate 
							fail("invalid avpkt argument for classless qdisc in configuration file. Using defaults\0", 5);
							break;
						}
						if(cf.cless_qdisc_opt == NULL) {
							cf.cless_qdisc_opt = (cless_qdisc_options *)malloc(sizeof(cless_qdisc_options));
							if(cf.cless_qdisc_opt == NULL)
								fail(strerror(errno),1);
							(void)memset(cf.cless_qdisc_opt, 0, sizeof(cless_qdisc_options));
						}
						cf.cless_qdisc_opt->red_avpkt = (size_t)atoi(token);
					}
					else if(!strncmp(token, "ecn", 3)) { //ecn 
						if(cf.cless_qdisc_opt == NULL) {
							cf.cless_qdisc_opt = (cless_qdisc_options *)malloc(sizeof(cless_qdisc_options));
							if(cf.cless_qdisc_opt == NULL)
								fail(strerror(errno),1);
							(void)memset(cf.cless_qdisc_opt, 0, sizeof(cless_qdisc_options));
						}
						cf.cless_qdisc_opt->red_ecn = 1;
					}
					else {
						fail("garbage after clessless qdisc argument in configuration file\0", 5);
						break;
					}
							
					token = strtok(NULL, delim); //get next token
				}
				continue;
			}
			else if(!strncmp(token, "team-members", 12)) {//process team members
				token = strtok(NULL, delim); //try to get the members
				if(token == NULL) {//missing qdisc
					fail("missing team members argument in configuration file. Using p2pwnc mode (default)\0", 5);
					continue;	
				}
				//update configuration file struct
				if(!strncmp(token, "local", 5))
					cf.team = TEAM;
				else if(!strncmp(token, "p2pwnc", 6))
					cf.team = NOT_TEAM;
				else {
					fail("invalid team members argument in configuration file. Using p2pwnc mode (default)\0", 5);
					continue;
				}
				token = strtok(NULL, delim);//check if there are any more tokens left
				if(token != NULL) {
					if(token[0] != '#')
						fail("garbage after team members argument in configuration file\0", 5);
				}
				continue;
			}
			else if(!strncmp(token, "connection-timeout", 18)) {//process timeout interval
				token = strtok(NULL, delim); //try to get timeout value
				if(token == NULL) {//missing
					fail("missing connection timeout argument in configuration file. Using default\0", 5);
					continue;
				}
				if(!atoi(token)) { //timeout argument number argument is invalid
					fail("invalid connection timeout argument in configuration file. Using default\0", 5); 
					continue;
				}
				//update configuration file struct
				cf.timeout = atoi(token);
				token = strtok(NULL, delim);//check if there are any more tokens left
				if(token != NULL) { 
					if(token[0] != '#')
						fail("garbage after connection timeout argument in configuration file\0", 5);
				}
				continue;
			}
			else if(!strncmp(token, "p2pwnc-server", 13)) {//process p2pwnc servers
				token = strtok(NULL, delim); //try to get the first server
				if(token == NULL) {//missing server list
					fail("missing p2pwnc server list in configuration file. Using default\0", 5);
					continue;
				}
				//update configuration file struct
				cf.serv_num = 0; 
				while(token != NULL && cf.serv_num < SERVER_LIST_MAX) {//process p2pwnc server list
					if(token[0] == '#')//comment
						break;
					//ip domain name validation 
					if(!inet_aton(token, &cf.servers[cf.serv_num])) {//ips are stored in network format (big-endian) 
						fail("invalid p2pwnc server ip in configuration file. Ignoring\0", 5);
						token = strtok(NULL, delim);
						continue;
					}
					token = strtok(NULL, delim);
					cf.serv_num++;
				}
				if(!cf.serv_num)
					fail("missing p2pwnc server list in configuration file. Using default\0", 2);
				continue;
			}
			else {
				if(token[0] != '#')
					fail("invalid option in configuration file. Skipping\0", 5);
				continue;
			}
		}
		
	}
	
	//validate for missing mandatory entries and fill with default values the optional entries
	
	//path it is definately defined
	if(cf.inet_dev == NULL || cf.p2pwnc_dev == NULL || !cf.ruplink || !cf.rdownlink)
		fail("missing mandatory entries in configuration file. Check inet-dev, p2pwnc-dev, real-uplink or real-downlink",  2); //mandatory entries must be filled
	
	//optional entries
	if(!cf.puplink) //default p2pwnc uplink = 50% real uplink
		cf.puplink = cf.ruplink / 2;
	else { //check if puplink > ruplink
		if(cf.puplink > cf.ruplink)	
			cf.puplink = cf.ruplink;
	}
	if(!cf.pdownlink) //default p2pwnc downlink = 50% real downlink
		cf.pdownlink = cf.rdownlink / 2;
	else {
		if(cf.pdownlink > cf.rdownlink)	
			cf.pdownlink = cf.rdownlink;
	}
	
	if(!cf.port)
		cf.port = htons(7777); //default port is 7777
	if(cf.tc == NULL) { //default tc path (/sbin/tc)
		cf.tc = (char *)malloc(9);
		if(cf.tc == NULL)
			fail(strerror(errno),1);
		(void)memset(cf.tc, '\0', 9);
		(void) strncpy(cf.tc, "/sbin/tc" , 8);
	}
	if(cf.iptables == NULL) { //default iptables path (/sbin/iptables)
		cf.iptables = (char *)malloc(15);
		if(cf.iptables == NULL)
			fail(strerror(errno),1);
		(void)memset(cf.iptables, '\0', 15);
		(void) strncpy(cf.iptables, "/sbin/iptables" , 14);
	}
	//no cf.cfull_qdisc check, cf.cless_qdisc cf.team check, 0 is default value - check only for RED qdisc
	if(cf.cless_qdisc == RED) {
		if(cf.cless_qdisc_opt == NULL) {
			cf.cless_qdisc_opt = (cless_qdisc_options *)malloc(sizeof(cless_qdisc_options));
			if(cf.cless_qdisc_opt == NULL)
				fail(strerror(errno),1);
			(void)memset(cf.cless_qdisc_opt, 0, sizeof(cless_qdisc_options));
			cf.cless_qdisc_opt->red_latency = RED_LATENCY;
			cf.cless_qdisc_opt->red_probability = RED_PROBABILITY;
			cf.cless_qdisc_opt->red_avpkt = RED_AVPKT;
			cf.cless_qdisc_opt->red_ecn = RED_ECN;
		}
		else {
			if(cf.cless_qdisc_opt->red_latency == 0)
				cf.cless_qdisc_opt->red_latency = RED_LATENCY;
			if(cf.cless_qdisc_opt->red_probability == 0)
				cf.cless_qdisc_opt->red_probability = RED_PROBABILITY;
			if(cf.cless_qdisc_opt->red_avpkt == 0)
				cf.cless_qdisc_opt->red_avpkt = RED_AVPKT;
		}
	}	
	if(!cf.timeout) //connections timeout interval
		cf.timeout = 5;
	if(!cf.serv_num) { //no p2pwnc server list
		//add only localhost (127.0.0.1) to list - ip is stored in network format (big-endian)
		inet_aton("127.0.0.1", &cf.servers[0]);
		cf.serv_num++;
	}

	//sort server list!
	qsort(cf.servers, cf.serv_num, sizeof(struct in_addr), in_addr_cmp);

	//report configuration files in syslog
	if(!strncmp(cf.path, "-", 1))
		syslog(LOG_FACILITY | LOG_INFO, "Settings loaded from stdin");	
 	else
		syslog(LOG_FACILITY | LOG_INFO, "Settings loaded using configuration file: %s",cf.path);	
	syslog(LOG_FACILITY | LOG_INFO, "inet-dev: %s",cf.inet_dev);	
 	syslog(LOG_FACILITY | LOG_INFO, "p2pwnc-dev: %s",cf.p2pwnc_dev);	
	syslog(LOG_FACILITY | LOG_INFO, "real-ulink: %dKbit",cf.ruplink);	
 	syslog(LOG_FACILITY | LOG_INFO, "real-downlink: %dKbit",cf.rdownlink);	
 	syslog(LOG_FACILITY | LOG_INFO, "p2pwnc-uplink: %dKbit",cf.puplink);	
 	syslog(LOG_FACILITY | LOG_INFO, "p2pwnc-downlink: %dKbit",cf.pdownlink);	
 	syslog(LOG_FACILITY | LOG_INFO, "port: %d",ntohs(cf.port));
	syslog(LOG_FACILITY | LOG_INFO, "tc-path: %s",cf.tc);	
 	syslog(LOG_FACILITY | LOG_INFO, "iptables-path: %s",cf.iptables);
	switch(cf.cfull_qdisc) {
		case HTB:
			if(cf.cfull_qdisc_opt != NULL) {
				if(cf.cfull_qdisc_opt->htb_r2q != 0 && cf.cfull_qdisc_opt->htb_quantum != 0)
					syslog(LOG_FACILITY | LOG_INFO, "cfull-qdisc: htb r2q %d quantum %d", cf.cfull_qdisc_opt->htb_r2q, cf.cfull_qdisc_opt->htb_quantum);
				else {
					if(cf.cfull_qdisc_opt->htb_r2q != 0)
						syslog(LOG_FACILITY | LOG_INFO, "cfull-qdisc: htb r2q %d", cf.cfull_qdisc_opt->htb_r2q);
					else
						syslog(LOG_FACILITY | LOG_INFO, "cfull-qdisc: htb quantum %d", cf.cfull_qdisc_opt->htb_quantum);
				}
			}
			else
 				syslog(LOG_FACILITY | LOG_INFO, "cfull-qdisc: htb");
			break;
	}
 	switch(cf.cless_qdisc) {
		case BFIFO:
			if(cf.cless_qdisc_opt != NULL)
				syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: bfifo limit %d bytes", cf.cless_qdisc_opt->fifo_limit);	
			else
				syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: bfifo");	
			break;
		case PFIFO:
			if(cf.cless_qdisc_opt != NULL)
				syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: pfifo limit %d packets", cf.cless_qdisc_opt->fifo_limit);	
			else
				syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: pfifo");	
			break;
		case SFQ:
			if(cf.cless_qdisc_opt != NULL) {
				if(cf.cless_qdisc_opt->sfq_perturb != 0 && cf.cless_qdisc_opt->sfq_quantum != 0 && cf.cless_qdisc_opt->fifo_limit != 0) 	
					syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: sfq perturb %d sec quantum %d bytes limit %d packets", cf.cless_qdisc_opt->sfq_perturb, cf.cless_qdisc_opt->sfq_quantum, cf.cless_qdisc_opt->fifo_limit);
				else {
					if(cf.cless_qdisc_opt->sfq_perturb !=0) { 
						if(cf.cless_qdisc_opt->sfq_quantum != 0)
							syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: sfq perturb %d sec quantum %d bytes", cf.cless_qdisc_opt->sfq_perturb, cf.cless_qdisc_opt->sfq_quantum);
						else {
							if(cf.cless_qdisc_opt->fifo_limit != 0)
								syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: sfq perturb %d sec limit %d packets", cf.cless_qdisc_opt->sfq_perturb, cf.cless_qdisc_opt->fifo_limit);
							else
								syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: sfq perturb %d sec", cf.cless_qdisc_opt->sfq_perturb);
						}
						break;
					}
					if(cf.cless_qdisc_opt->sfq_quantum !=0) { 
						if(cf.cless_qdisc_opt->sfq_perturb != 0)
							syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: sfq perturb %d sec quantum %d bytes", cf.cless_qdisc_opt->sfq_perturb, cf.cless_qdisc_opt->sfq_quantum);
						else {
							if(cf.cless_qdisc_opt->fifo_limit != 0)
								syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: sfq quantum %d bytes limit %d packets", cf.cless_qdisc_opt->sfq_quantum, cf.cless_qdisc_opt->fifo_limit);
							else
								syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: sfq quantum %d bytes", cf.cless_qdisc_opt->sfq_quantum);
						}
						break;
					}
					if(cf.cless_qdisc_opt->fifo_limit !=0) { 
						if(cf.cless_qdisc_opt->sfq_perturb != 0)
							syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: sfq perturb %d sec limit %d packets", cf.cless_qdisc_opt->sfq_perturb, cf.cless_qdisc_opt->fifo_limit);
						else {
							if(cf.cless_qdisc_opt->sfq_quantum != 0)
								syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: sfq quantum %d bytes limit %d packets", cf.cless_qdisc_opt->sfq_quantum, cf.cless_qdisc_opt->fifo_limit);
							else
								syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: sfq limit %d packets", cf.cless_qdisc_opt->fifo_limit);
						}
						break;
					}
				}
			}
			else
				syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: sfq");	
			break;
		case RED:
			
			if(cf.cless_qdisc_opt->red_ecn == 0)
				syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: red latency %d ms probability %f avpkt %d bytes", cf.cless_qdisc_opt->red_latency, cf.cless_qdisc_opt->red_probability, cf.cless_qdisc_opt->red_avpkt);
			else
				syslog(LOG_FACILITY | LOG_INFO, "cless-qdisc: red latency %d ms probability %f avpkt %d bytes ecn", cf.cless_qdisc_opt->red_latency, cf.cless_qdisc_opt->red_probability, cf.cless_qdisc_opt->red_avpkt);
			break;
	}
	switch(cf.team)	{
		case TEAM:
 			syslog(LOG_FACILITY | LOG_INFO, "team-members: local");
			break;
		case NOT_TEAM:
 			syslog(LOG_FACILITY | LOG_INFO, "team-members: p2pwnc");	
			break;
	}
	syslog(LOG_FACILITY | LOG_INFO, "connection-timeout: %d sec", cf.timeout);
 	for(i=0;i<cf.serv_num;i++)
		syslog(LOG_FACILITY | LOG_INFO, "p2pwnc-server: %s", inet_ntoa(cf.servers[i]));
	
	// Try to close the file stream
	if(fclose(fp) != 0)
		fail(strerror(errno),4);
}
//################################################################

//################################################################
/*
 * in_addr comparator
 * 0 = lhs, rhs are equal
 * >0 = lhs > rhs
 * <0 lhs < rhs
 */
int in_addr_cmp(struct in_addr* lhs, struct in_addr* rhs) {
	return (int)lhs->s_addr - rhs->s_addr;
}
//################################################################

//################################################################
/* 
 * Set program's name to a valid 2..9 character name as POSIX 
 * conventions require 
 */
void set_progname(char* pname) {
	//use the global variables
	extern size_t PROGNAME_LEN;
	extern char progname[];

	//Ignore path prefix and store only program's name
	char* ppname = strrchr(pname,'/'); 
	if(ppname == NULL) {

		PROGNAME_LEN = strlen(pname);
		if(PROGNAME_LEN > PROGNAME_LEN_MAX-1)
			PROGNAME_LEN = PROGNAME_LEN_MAX-1;
		//init and copy
		memset(progname, '\0', PROGNAME_LEN+1);
		(void) strncpy(progname, pname, PROGNAME_LEN);
	}
	else {
		//same as the above
		PROGNAME_LEN = strlen(ppname+1); //+1 to ignore the first backslash in program's name
		if(PROGNAME_LEN > PROGNAME_LEN_MAX-1)
			PROGNAME_LEN = PROGNAME_LEN_MAX-1;
		memset(progname, '\0', PROGNAME_LEN+1);
		(void) strncpy(progname, ppname+1, PROGNAME_LEN);
	}
}
//################################################################

//################################################################
/* 
 * Print help and usage info 
 */ 
void usage(void) {
	//use the global variables
	extern char progname[];
	
	(void) fprintf(stderr, "\nUsage: %s ",progname);
	(void) fprintf(stderr, "[-h/--help] ");
	(void) fprintf(stderr, "[-v/--version] ");
	(void) fprintf(stderr, "[-c/--conf file]\n\n");
	(void) fprintf(stderr, "\t-h --help\t\tDisplay this message\n");
	(void) fprintf(stderr, "\t-v --version\t\tDisplay version\n");
	(void) fprintf(stderr, "\t-c --conf\t\tTraffic Controller\'s configuration file\n\n");
}
//################################################################

//################################################################
/*
 * Main 
 * - parse command line options
 * - set a valid program name - POSIX [2..9]
 * - register deallocator exit function
 * - fill configuration file
 * - generate chile process
 * - exit
 * -- [child]
 * --
 * -- set init as parent process
 * -- become session leader and group leader
 * -- close stdin, stdout, stderr
 * -- change root dir to /
 * -- clear mask creation mode
 * -- increase process priority
 * -- start TC server
 */
int main(int argc,char** argv) {
 //use the global variables
 extern char progname[];
 extern const char* version;
 extern FILE* pidfile;
 
 // POSIX getopt() 
 int oc; //option char (h,c,v...)
 extern char* optarg; //option argument (for -c file, optarg = file);
 extern int optind, opterr, optopt; //getopt misc external variables
 opterr = 0; //turn off error messages
 
 // getopt_long options 
 struct option longopts[] = {
 	{"conf",	required_argument,	NULL,	'c'}, //conf equals c
	{"pid",		required_argument,	NULL,	'p'}, //pid equals p
 	{"help",	no_argument,		NULL,	'h'}, //help equals h
 	{"version",	no_argument,		NULL,	'v'}, //version equals v
 	{0, 0, 0, 0}
 };
 
 // Error messages buffer 
 char errmsg[MSGBUF_MAX]; 
 
 // Process id
 pid_t pid;
 
 // Set to progname a valid [2..9] character name  - POSIX 
 set_progname(argv[0]);

 // Init configuration file stracture - replace the obsolete bzero 
 (void)memset(&cf, 0, sizeof(config_file));	
 // Init deallocator - registration is active only for normal termination
 (void)atexit(dealloc);
 
 // Init syslog
 openlog(progname, LOG_PID, LOG_FACILITY);
 syslog(LOG_FACILITY | LOG_INFO, "Starting P2PWNC Traffic Controller %s", version);
 
 // Start parsing only if we have arguments 
 if(argc > 1) {
 
 	// Parse comand line options -c/--conf -p/--pid -h/--help and -v/--version 
 	while((oc = getopt_long(argc, argv, ":c:p:hv", longopts, NULL)) != -1) {
		switch(oc) {
		case 'c': 
			// Path to the config file 
			if (cf.path != NULL) //clean from previous allocations
				free(cf.path);
			cf.path = (char *)malloc(strlen(optarg)+1);
			if(cf.path == NULL)
				fail(strerror(errno),1);
			(void)memset(cf.path, '\0', strlen(optarg)+1);
			(void)strncpy(cf.path,optarg,strlen(optarg));
			break;
		case 'p':
			//Create pid file 
			//clear from previous allocations
			if(pidfilename != NULL)
				free(pidfilename);
			pidfilename = (char *)malloc(strlen(optarg)+1);
			if(pidfilename == NULL)
				fail(strerror(errno),1);
			(void)memset(pidfilename, '\0', strlen(optarg)+1);
			(void)strncpy(pidfilename,optarg,strlen(optarg));
			if((pidfile = fopen(pidfilename,"w")) == NULL)
				fail("Failed to open pid file",4);
			break;
		case 'h':
			// Display usage info 
			usage();
 			fail(NULL,-1); // Normal termination 
		case 'v': 
			// Display version info 
			(void)fprintf(stderr, "Version: %s %s\n",progname,version);
 			fail(NULL,-1); // Normal termination 
		case ':':
			// Missing argument 
			(void)memset(errmsg, '\0', MSGBUF_MAX);
			(void)sprintf(errmsg,"option -%c requires an argument",optopt);
			fail(errmsg,2);
		case '?':
		default : 
			// Invalid argument 
			(void)memset(errmsg, '\0', MSGBUF_MAX);
			(void)sprintf(errmsg,"option -%c is invalid",optopt);
			fail(errmsg,2);
		}
 	}
 }
 //If path is not set init to deafult config path /etc/p2pwnc/tc.conf 
 if (cf.path == NULL) {
 	cf.path = (char *)malloc(strlen(CONFILE)+1);	
 	if(cf.path == NULL)
 		fail(strerror(errno),1);
 	(void)memset(cf.path, '\0', strlen(CONFILE));
 	(void)strncpy(cf.path,CONFILE,strlen(CONFILE));
 }
 
 //Default pid file 
 if(pidfilename == NULL) {
	pidfilename = (char *)malloc(strlen(PIDFILE)+1);
	if(pidfilename == NULL)
		fail(strerror(errno),1);
	(void)memset(pidfilename, '\0', strlen(PIDFILE));
	(void)strncpy(pidfilename,PIDFILE,strlen(PIDFILE));
 	if((pidfile = fopen(pidfilename, "w"))== NULL)//fail
		fail(strerror(errno),1);
 }
 
 //Fill configuration file stracture from file or stdin
 read_config(cf.path);
 
 //Become a daemon process
 if((pid = fork()) < 0)
 	fail(strerror(errno),1);
 else if(!pid) { //child process
	//1) Set init as parent process - parent process returns
 	//2) Become session leader and group leader
 	//3) No controlling terminal
 	//4) Change root dir to /
 	//5) Clear mask creation mode
	//6) Increase process priority
	//7) Start TC server

 	(void)setsid(); //become session leader
	//no controlling terminal
	(void)fclose(stdin);
	(void)fclose(stdout);
	(void)fclose(stderr);
 	(void)chdir("/"); //change working directory
 	(void)umask(0); //clear file mode creation mask
	(void)nice(-15); //change prio
	
	//write child pid in pid file
	(void)fprintf(pidfile,"%d",getpid());
	(void)fflush(pidfile);
	(void)fclose(pidfile); //close pid file
	
	//server start
	struct sigaction sigac; //signal handling
	struct sockaddr_in server; //server's address/port
	int sfd; //socket descriptor
	
	(void)sigfillset(&sigac.sa_mask); //block all signals when signal handler function run
	sigac.sa_flags = 0;
	sigac.sa_handler = terminate; //signal handler
	
	//signals - call terminate to cleanup and shutdown
	(void)sigaction(SIGHUP,&sigac,NULL);
	(void)sigaction(SIGINT,&sigac,NULL);
	(void)sigaction(SIGQUIT,&sigac,NULL);
	(void)sigaction(SIGABRT,&sigac,NULL);
	(void)sigaction(SIGTERM,&sigac,NULL);
	
	// TCP/IP socket - RFCs: 791, 793, 1112, 1122, 1191, 1323, 2001, 2018, 2481, 2581, 2883
						   //Enabled options// 
						   //NewReno - F,S,DS ACK
						   //Available options - Use sysctl to enable them /proc/sys/net/ipv4/ //
						   //High performance extensions
						   //BIC, Westwood+, Vegas congestion control support
						   //F-RTO algorithm
						   //ECN
	if((sfd = socket (AF_INET, SOCK_STREAM, 0) )== -1) {//socket failed
		(void)memset(errmsg, '\0', MSGBUF_MAX);
		(void)sprintf(errmsg, "socket initialization failed: %s", strerror(errno));
		fail(errmsg, 7);
	}
	
	(void)memset(&server, 0, sizeof(struct sockaddr_in)); //init address/port
	server.sin_family = AF_INET;
	server.sin_port = cf.port;
	server.sin_addr.s_addr = htonl(INADDR_ANY);
	
	if(bind(sfd, (struct sockaddr *)&server, sizeof(server)) == -1) {//bind TCP/IP socket to the specified address/port
		(void)memset(errmsg, '\0', MSGBUF_MAX);
		(void)sprintf(errmsg, "binding failed: %s", strerror(errno));
		fail(errmsg, 7);
	}
	if(listen(sfd, 10) == -1) {//accept connections
		(void)memset(errmsg, '\0', MSGBUF_MAX);
		(void)sprintf(errmsg, "failed while accepting connections: %s", strerror(errno));
		fail(errmsg, 7);
	}
	
	tc_server(sfd); //start TC server
	
	(void)close(sfd); //close the socket connection
	fail(NULL, -1); //Normal termination for daemon process - Never reached tc_server runs forever
 }
 else //parent goes bye - bye :)
 	fail(NULL,0); //Normal termination for not daemon process
	
 return 0; //Never reached, just for fun :)
}
