/**
* p2pwnc-tc.h - P2PWNC Traffic Controller module
*
* Copyright (C) 2005-2006 Vasileios P. Kemerlis
*	
* Author:  Vasileios P. Kemerlis
* Contact: vpk@cs.aueb.gr
*          hdkiller@venus.cslab.aueb.gr
*          http://www.mm.aueb.gr/~vpk/
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*
*/


#ifndef _P2PWNC_TC_H
#define _P2PWNC_TC_H

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <netdb.h>
#include <net/if.h>
#include <netinet/in.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <syslog.h>
#include <unistd.h>

/* Global Variables and Structures */
//Max Lengths
#define PROGNAME_LEN_MAX 10
#define SERVER_LIST_MAX 20
#define MSGBUF_MAX 100
/*Constants*/
//Tc
#define CLEAR 0
#define CLEAR_INIT 1
//Timer 
#define ACTIVATE_TIMER 0
#define DEACTIVATE_TIMER 1
//Team member
#define TEAM 1
#define NOT_TEAM  0
//Classfull qdisc
#define HTB 0
//Classless qdisc
#define BFIFO 0
#define PFIFO 1
#define SFQ 2
//Red with default options
#define RED 3
#define RED_LATENCY 500
#define RED_AVPKT 1000
#define RED_PROBABILITY 0.02
#define RED_ECN 0
//Syslog facility
#define LOG_FACILITY LOG_LOCAL0
//default paths
#define PIDFILE "/var/run/p2pwnc/tc.pid"
#define CONFILE "/etc/p2pwnc/tc.conf"
#define TMPFILE "/tmp/tc"

// POSIX program name. Up to 9 chars 
size_t PROGNAME_LEN;
char progname[PROGNAME_LEN_MAX];

// Program Version 
const char* version = "1.0alpha - 02/05/2006\0";

// Client - P2PWNC Roamer details
typedef struct p2pwn_roamer {
	struct in_addr ip_address; //ip-address
	uint8_t team_member; // 1 if client is a team member, 0 if not
	uint8_t qos_level; //qos type of service, available values are 1-255 
	void* next; //next client (one-way connected list)
} p2pwnc_client;

// Resource allocation pool
typedef struct tc_client_list {
	p2pwnc_client* cls; //clients list
	size_t num; //clients counter
	size_t qos_level_sum; //sum of all qos_levels
} resource_pool;

// Class full qdisc options - r2q and quantum for htb 
typedef struct classfull_qdisc_options {
	uint8_t htb_r2q; //htb r2q value 
	size_t htb_quantum; //htb quantum value 
} cfull_qdisc_options;

// Class less qdisc options - limit for p|bfifo and quantum, perturb, limit for sfq
typedef struct classless_qdisc_options {
	size_t fifo_limit; // bfifo queue size in bytes or pfifo queue size in packets or sfq queue size in packets
	uint8_t sfq_perturb; //sfq rehashing time 
	size_t sfq_quantum; //sfq quantum value 
	size_t red_latency; //red desired latency
	float red_probability; //red probability
	size_t red_avpkt; //red average packet size
	uint8_t red_ecn; //red ecn option
} cless_qdisc_options;

// Config file representation 
typedef struct cfile {
	//Mandatory
	char* path; //Path to config file
	char* inet_dev; //Internet gateway device 
	char* p2pwnc_dev; //P2PWNC dev
	size_t ruplink; //Internet iface uplink speed in Kbps
	size_t rdownlink; //Internet iface downlink speed in Kbps
	
	//Optional
	size_t puplink; //P2PWNC max uplink speed in Kbps
	size_t pdownlink; //P2PWNC max downlink speed in Kbps
	unsigned short int port; //Traffic Controller's port
	char* tc; //Path to tc (iproute2) exec
	char* iptables; //Path to iptables exec
	uint8_t cfull_qdisc; //The classfull qdisc to use (htb = 0)
	cfull_qdisc_options* cfull_qdisc_opt; //Advanced options for the classfull qdisc
	uint8_t cless_qdisc; //The classless qdisc to use (bfifo = 0 | pfifo = 1 | sfq = 2 | red = 3)
	cless_qdisc_options* cless_qdisc_opt; //Advanced options for the classless qdiscs
	uint8_t team; //Team members flag (local = 1 | p2pwnc = 0)
	unsigned short int timeout; //connections time out - seconds
	struct in_addr servers[SERVER_LIST_MAX]; //List with p2pwnc servers (10.0.0.1 10.0.0.2 ...)
	uint8_t serv_num; //Servers count
} config_file;

// Configuration file
config_file cf;

// Resource pool - Details about each allocation 
resource_pool rpool;

int cfd; //file descriptor for accepted connections (client)- global in order to be accessed from signal handlers
FILE* pidfile; //pid file struct
char* pidfilename; //pid file name
volatile sig_atomic_t alarm_sig; //atomic flag for alarm (alarm sets variable as true)

/* Function prototypes */

//handlers
void set_timer(struct sigaction* alrm, uint8_t action);
void fail(char* reason, int retvalue);
void terminate(int sig);
void close_connection(int sig);

//basic
void dealloc(void);
int in_addr_cmp(struct in_addr* lhs, struct in_addr* rhs); 
void read_config(char* path); 
void set_progname(char* pname); 
void usage(void);

//tc server related
uint8_t init_tc(uint8_t mode, struct in_addr* address);
uint8_t update_resource_pool(struct in_addr address, uint8_t member, uint8_t qlevel);
void print_updated_tc(void);
uint8_t is_valid_header(char* msg); 
uint8_t is_valid_content(uint8_t length, char* msg, struct in_addr* address, uint8_t* team, uint8_t* qlevel); 
void tc_server(int fd); 

#endif 
